/**
* Tuex v0.3.16
* (c) 2018 Alexey Iskhakov
* @license MIT
*/


const desc = Object.getOwnPropertyDescriptor;
const keysOf = Object.getOwnPropertyNames;
const isObject = (descriptor) => {
    const obj = descriptor.value;
    if (!!obj)
        return Object.prototype.toString.apply(obj) === '[object Object]';
    else
        return false;
};
const isFunction = (fn) => {
    return !!fn && (fn instanceof Function || typeof fn === 'function');
};

const isValue = (descriptor) => {
    return !!descriptor && ((!isFunction(descriptor.value) && !descriptor.get && !descriptor.set)
        || (!descriptor.value && descriptor.get && descriptor.set));
};
const isGetter = (descriptor) => {
    return !!descriptor && !!descriptor.get && !descriptor.set;
};
const isSetter = (descriptor) => {
    return !!descriptor && !descriptor.get && !!descriptor.set;
};
const error = (message) => {
    if (process && process.env.NODE_ENV !== 'production') {
        console.error(message);
    }
};

let _vue;
/**
 * Store
 *
 * @class Store
 * @template T
 */
class Store {
    /**
     * Creates an instance of Store.
     * @param {(T | (new () => T) | (() => T))} target - can be a plain object, function that returns an object or a constructor function (class)
     * @param {{
     *       strict?: boolean, // - whether to disallow state modifications
     *       plugins?: ((this: Store<T>) => any)[], // - optional plugins to install
     *     }} options
     * @memberof Store
     */
    constructor(target, options) {
        this._eventPool = {
            value: [],
            getter: [],
            setter: [],
            action: [],
            global: []
        };
        this._strict = false;
        this._store = { state: null };
        const { strict, plugins } = options || { strict: false, plugins: [] };
        this._strict = strict;
        this.replaceStore(target);
        plugins && plugins.forEach(plugin => plugin.apply(this));
    }
    _storeEvent(type, store, key, ...args) {
        this._eventPool[type].forEach(callback => callback(store, key, ...args));
    }
    get store() {
        return this._store.state;
    }
    set store(value) {
        error(`Can't assign ${value} to store:
Explicit store assignment is prohibited! Consider using [replaceStore] instead!`);
    }
    /**
     * Subscribe for store events.
     * Callback is executed BEFORE the event!
     *
     * @param {'value' | 'getter' | 'setter' | 'action' | 'global'} type
     * @param {(store: T, key: string) => any} callback
     * @returns a funciton to unsubscribe from event
     * @memberof Tuex
     */
    subscribe(type, callback) {
        this._eventPool[type].push(callback);
        return () => {
            this._eventPool[type] = [...this._eventPool[type].filter(c => c != callback)];
        };
    }
    /** replaceStore
     *
     * A function that replaces the current store with the other one,
     * converting it from a target object/function/constructor
     *
     * @param {(T | (new () => T) | (() => T))} target
     * @memberof Tuex
     */
    replaceStore(target) {
        let plain;
        if (isFunction(target)) {
            try {
                plain = new target();
            }
            catch (e) {
                plain = target();
            }
            this._store.state = this.objectToStore(plain, target);
        }
        else {
            plain = target;
            this._store.state = this.objectToStore(plain);
        }
        _vue && (_vue.prototype.$store = this.store);
    }
    /** objectToStore
     *
     * Converts a plain js object into a valid Tuex-store
     *
     * @param {T} plain - object to convert
     * @param {new () => T} [constructor] - constructor (if any)
     * @returns {T} - converted store
     * @memberof Tuex
     */
    objectToStore(plain, constructor) {
        const obj = {};
        const keys = keysOf(plain);
        if (isFunction(constructor))
            keys.push(...keysOf(constructor.prototype));
        for (let key of keys) {
            const define = (prop) => Object.defineProperty(obj, key, prop);
            const descriptor = desc(plain, key) || desc(constructor.prototype, key);
            const callStoreEvent = (type, ...args) => this._storeEvent.call(this, type, plain, key, ...args);
            if (isFunction(plain[key]))
                define({
                    value(...args) {
                        callStoreEvent('action', ...args);
                        return plain[key].apply(obj, args);
                    }
                });
            else if (isValue(descriptor)) {
                const isKeyObject = isObject(descriptor);
                let get, set;
                if (isKeyObject) {
                    plain[key] = this.objectToStore(plain[key]);
                }
                if (descriptor.get)
                    get = () => {
                        callStoreEvent('value');
                        // Should replace getter's context of 'this'
                        return descriptor.get.call(obj);
                    };
                else
                    get = () => {
                        callStoreEvent('value');
                        return plain[key];
                    };
                if (descriptor.set && !this._strict)
                    set = value => {
                        callStoreEvent('global', value);
                        callStoreEvent('value', value);
                        // Should replace setter's context of 'this'
                        descriptor.set.call(obj, value);
                    };
                else if (!this._strict)
                    set = value => {
                        callStoreEvent('global', value);
                        callStoreEvent('value', value);
                        plain[key] = isKeyObject ? this.objectToStore(value) : value;
                    };
                else
                    set = () => error('Explicit mutations of store values are prohibited!\nPlease, use setters instead or disable the [strict] flag!');
                define({ enumerable: true, get, set });
            }
            else if (isGetter(descriptor))
                define({
                    get: () => {
                        callStoreEvent('getter');
                        // Should replace getter's context of 'this'
                        return descriptor.get.call(obj);
                    }
                });
            else if (isSetter(descriptor))
                define({
                    set: value => {
                        callStoreEvent('global', value);
                        callStoreEvent('setter', value);
                        // Should replace setter's context of 'this'
                        descriptor.set.call(obj, value);
                    }
                });
            else
                error('Descriptor of ' + key + ' is wrong!');
        }
        return Object.seal(obj);
    }
}
function install(Vue) {
    if (_vue && Vue === _vue) {
        error('[tuex] is already installed. Vue.use(Tuex) should be called only once.');
        return;
    }
    _vue = Vue;
}


var Tuex$1 = Object.freeze({
	Store: Store,
	install: install
});

export { Store };
export default Tuex$1;
